import tkinter as tk
from typing import Tuple, Union, Optional
import sys 

_root: Optional[tk.Tk] = None
_canvas: Optional[tk.Canvas] = None
_pixels = {}

def init(width: int = 320, height: int = 240, title: str = "Kandinsky") -> None:
    global _root, _canvas
    _root = tk.Tk()
    _root.title(title)
    _canvas = tk.Canvas(_root, width=width, height=height, bg="black")
    _canvas.winfo_toplevel().protocol("WM_DELETE_WINDOW", sys.exit)
    _canvas.pack()



_current_color = (0, 0, 0)


def color(
    r: Union[int, float, Tuple[int, int, int]], g: int = 0, b: int = 0
) -> Tuple[int, int, int]:
    """Defines the color from the values of r,g,b.

    Args:
        r (Union[int, float, Tuple[int, int, int]]): Red component or full color tuple
        g (int, optional): Green component. Defaults to 0.
        b (int, optional): Blue component. Defaults to 0.

    Returns:
        Tuple[int, int, int]: Color tuple.
    """
    global _current_color
    if isinstance(r, tuple):
        _current_color = r
    else:
        _current_color = (r, g, b)
    return _current_color


def get_pixel(x: int, y: int) -> Tuple[int, int, int]:
    """Returns the pixel x,y color as a tuple (r,g,b).

    Args:
        x (int): X-coordinate
        y (int): Y-coordinate

    Returns:
        Tuple[int, int, int]: Pixel at that location
    """
    return _pixels.get((x, y), (0, 0, 0))


def set_pixel(x: int, y: int, col: Tuple[int, int, int] = None) -> None:
    """Colors the pixel x,y of the color color.

    Args:
        x (int): X-coordinate
        y (int): Y-coordinate
        color (Tuple[int, int, int], optional): Color tuple. Defaults to current color.
    """
    if col is None:
        col = _current_color
    hex_color = "#%02x%02x%02x" % col
    _pixels[(x, y)] = col
    _canvas.create_line(x, y, x + 1, y, fill=hex_color)


def draw_string(
    text: str,
    x: int,
    y: int,
    col1: Tuple[int, int, int] = (255, 255, 255),
    col2: Tuple[int, int, int] = None,
) -> None:
    """
    Displays text from the pixel x,y

    Args:
        text (str): Text to display
        x (int): X-coordinate
        y (int): Y-coordinate
        col1 (tuple[int,int,int], optional): Foreground pixel. Defaults to white.
        col2 (tuple[int,int,int], optional): Background pixel. Defaults to None (transparent).
    """

    # cant really put a background using tkinter but might find a way for it!
    fg = "#%02x%02x%02x" % col1
    if col2:
        bg = "#%02x%02x%02x" % col2
    else:
        bg = ""
    _canvas.create_text(x, y, text=text, fill=fg, anchor="nw", font=("Arial", 12))


def fill_rect(
    x: int, y: int, width: int, height: int, pixel: Tuple[int, int, int]
) -> None:
    """
    Fills a rectangle at pixel (x,y) with the color color.

    Args:
        x (int): X-coordinate
        y (int): Y-coordinate
        width (int): Width of the rectangle
        height (int): Height of the rectangle
        pixel (tuple[int,int,int]): Pixel color
    """
    hex_color = "#%02x%02x%02x" % pixel
    _canvas.create_rectangle(
        x, y, x + width, y + height, outline=hex_color, fill=hex_color
    )
    for i in range(x, x + width):
        for j in range(y, y + height):
            _pixels[(i, j)] = pixel

def done() -> None:
    """Start the Tkinter main loop to keep the window open."""
    if _root:
        _root.mainloop()


from enum import IntEnum
import keyboard


class _Key(IntEnum):
    LEFT = 0
    UP = 1
    DOWN = 2
    RIGHT = 3
    OK = 4
    BACK = 5
    HOME = 6
    ONOFF = 7
    SHIFT = 12
    ALPHA = 13
    XNT = 14
    VAR = 15
    TOOLBOX = 16
    BACKSPACE = 17
    EXP = 18
    LN = 19
    LOG = 20
    IMAGINARY = 21
    COMMA = 22
    POWER = 23
    SINE = 24
    COSINE = 25
    TANGENT = 26
    PI = 27
    SQRT = 28
    SQUARE = 29
    SEVEN = 30
    EIGHT = 31
    NINE = 32
    LEFTPARENTHESIS = 33
    RIGHTPARENTHESIS = 34
    FOUR = 36
    FIVE = 37
    SIX = 38
    MULTIPLICATION = 39
    DIVISION = 40
    ONE = 42
    TWO = 43
    THREE = 44
    PLUS = 45
    MINUS = 46
    ZERO = 48
    DOT = 49
    EE = 50
    ANS = 51
    EXE = 52

_KEYMAP = {
    _Key.LEFT: "left",
    _Key.UP: "up",
    _Key.DOWN: "down",
    _Key.RIGHT: "right",
    _Key.OK: "enter",
    _Key.BACK: "backspace",
    _Key.HOME: "home",
    _Key.ONOFF: "esc",
    _Key.SHIFT: "shift",
    _Key.ALPHA: "caps lock",
    _Key.XNT: "x",
    _Key.VAR: "v",
    _Key.TOOLBOX: "t",
    _Key.BACKSPACE: "backspace",
    _Key.EXP: "e",
    _Key.LN: "l",
    _Key.LOG: "g",
    _Key.IMAGINARY: "i",
    _Key.COMMA: ",",
    _Key.POWER: "^",
    _Key.SINE: "s",
    _Key.COSINE: "c",
    _Key.TANGENT: "t",
    _Key.PI: "p",
    _Key.SQRT: "r",
    _Key.SQUARE: "q",
    _Key.SEVEN: "7",
    _Key.EIGHT: "8",
    _Key.NINE: "9",
    _Key.LEFTPARENTHESIS: "(",
    _Key.RIGHTPARENTHESIS: ")",
    _Key.FOUR: "4",
    _Key.FIVE: "5",
    _Key.SIX: "6",
    _Key.MULTIPLICATION: "*",
    _Key.DIVISION: "/",
    _Key.ONE: "1",
    _Key.TWO: "2",
    _Key.THREE: "3",
    _Key.PLUS: "+",
    _Key.MINUS: "-",
    _Key.ZERO: "0",
    _Key.DOT: ".",
    _Key.EE: "e",
    _Key.ANS: "a",
    _Key.EXE: "enter",
}


def keydown(k: _Key) -> bool:
    """Returns True if the k key in argument is pressed and False otherwise.


    Args:
        k (Key): A key constant from the ion module


    Raises:
        ValueError: If the provided key constant is not recognized.


    Returns:
        bool: True if the corresponding computer key is pressed,
        False otherwise.
    """
    if not k in _KEYMAP:
        raise ValueError(f"Unknown key {k}")
    return keyboard.is_pressed(_KEYMAP[k])

KEY_LEFT = _Key.LEFT
KEY_UP = _Key.UP
KEY_DOWN = _Key.DOWN
KEY_RIGHT = _Key.RIGHT
KEY_OK = _Key.OK
KEY_BACK = _Key.BACK
KEY_HOME = _Key.HOME
KEY_ONOFF = _Key.ONOFF
KEY_SHIFT = _Key.SHIFT
KEY_ALPHA = _Key.ALPHA
KEY_XNT = _Key.XNT
KEY_VAR = _Key.VAR
KEY_TOOLBOX = _Key.TOOLBOX
KEY_BACKSPACE = _Key.BACKSPACE
KEY_EXP = _Key.EXP
KEY_LN = _Key.LN
KEY_LOG = _Key.LOG
KEY_IMAGINARY = _Key.IMAGINARY
KEY_COMMA = _Key.COMMA
KEY_POWER = _Key.POWER
KEY_SINE = _Key.SINE
KEY_COSINE = _Key.COSINE
KEY_TANGENT = _Key.TANGENT
KEY_PI = _Key.PI
KEY_SQRT = _Key.SQRT
KEY_SQUARE = _Key.SQUARE
KEY_SEVEN = _Key.SEVEN
KEY_EIGHT = _Key.EIGHT
KEY_NINE = _Key.NINE
KEY_LEFTPARENTHESIS = _Key.LEFTPARENTHESIS
KEY_RIGHTPARENTHESIS = _Key.RIGHTPARENTHESIS
KEY_FOUR = _Key.FOUR
KEY_FIVE = _Key.FIVE
KEY_SIX = _Key.SIX
KEY_MULTIPLICATION = _Key.MULTIPLICATION
KEY_DIVISION = _Key.DIVISION
KEY_ONE = _Key.ONE
KEY_TWO = _Key.TWO
KEY_THREE = _Key.THREE
KEY_PLUS = _Key.PLUS
KEY_MINUS = _Key.MINUS
KEY_ZERO = _Key.ZERO
KEY_DOT = _Key.DOT
KEY_EE = _Key.EE
KEY_ANS = _Key.ANS
KEY_EXE = _Key.EXE

__all__ = [
    "KEY_LEFT", "KEY_UP", "KEY_DOWN", "KEY_RIGHT",
    "KEY_OK", "KEY_BACK", "KEY_HOME", "KEY_ONOFF",
    "KEY_SHIFT", "KEY_ALPHA", "KEY_XNT", "KEY_VAR", "KEY_TOOLBOX",
    "KEY_BACKSPACE", "KEY_EXP", "KEY_LN", "KEY_LOG", "KEY_IMAGINARY",
    "KEY_COMMA", "KEY_POWER", "KEY_SINE", "KEY_COSINE", "KEY_TANGENT",
    "KEY_PI", "KEY_SQRT", "KEY_SQUARE",
    "KEY_SEVEN", "KEY_EIGHT", "KEY_NINE", "KEY_LEFTPARENTHESIS", "KEY_RIGHTPARENTHESIS",
    "KEY_FOUR", "KEY_FIVE", "KEY_SIX", "KEY_MULTIPLICATION", "KEY_DIVISION",
    "KEY_ONE", "KEY_TWO", "KEY_THREE", "KEY_PLUS", "KEY_MINUS",
    "KEY_ZERO", "KEY_DOT", "KEY_EE", "KEY_ANS", "KEY_EXE",
    "keydown"
]




from typing import List, Any
import random as _random;

def getrandbits(k: int) -> int:
    """Returns an integer with k random bits.

    Args:
        k (int): Number of random bits to generate. Must be a non-negative integer.

    Returns:
        int: An integer in the range [0, 2**k), i.e., with k bits of randomness.
    """    
    return _random.getrandbits(k)

def seed(x: int) -> None:
    """Initialize the random number generator.

    Args:
        x (int): Seed value to initialize the generator.
    """    
    return _random.seed(x)

def randrange(start: int, stop: int = None) -> int:
    """Returns a random number in range(start,stop)

    Args:
        start (int): Start of range if stop is specified, else stop value
        stop (int, optional): End of range (exclusive). Defaults to None.

    Returns:
        int: Random integer in the specified range.
    """    
    return _random.randrange(start, stop)

def randint(a: int, b: int) -> int:
    """Returns an integer in [a,b].

    Args:
        a (int): Lower bound (inclusive).
        b (int): Upper bound (inclusive).

    Returns:
        int: Random integer in [a, b].
    """    
    return _random.randint(a, b)

def choice(list: List[Any]) -> Any:
    """Returns a random number in the list.

    Args:
        list (List[Any]): Non-empty list to choose from.

    Raises:
        IndexError: If the list is empty.
        
    Returns:
        Any: Random element from the list.
    """    
    if not list:
        raise IndexError("Cannot choose from empty list")
    return _random.choice(list)

def random() -> float:
    """Returns a random floating point number in [0,1).

    Returns:
        float: A random float x such that 0 <= x < 1.
    """    
    return _random.random()

def uniform(a: float, b: float) -> float:
    """Returns a random floating point number in [a,b].

    Args:
        a (float): Lower bound of the range.
        b (float): Upper bound of the range.

    Returns:
        float: A random float x such that a <= x <= b.
    """    
    return _random.uniform(a, b)

__all__ = [
    "getrandbits",
    "seed",
    "randrange",
    "randint",
    "choice",
    "random",
    "uniform"
]


from typing import List, Tuple, Union
import numpy as _numpy
from numpy.typing import ArrayLike, NDArray


def array(lst: List) -> NDArray:
    """Creates an array from a list.

    Args:
        lst (List): Input list of numbers or nested lists.

    Returns:
        NDArray: NumPy array created from the given list.
    """
    return _numpy.array(lst)


def arange(start: int, stop: int) -> NDArray:
    """Returns an array of integers from start to stop-1.

    Args:
        start (int): Starting value.
        stop (int): End value (exclusive).

    Returns:
        NDArray: Array of integers.
    """
    return _numpy.arange(start, stop)


def concatenate(a: Tuple[ArrayLike, ArrayLike]) -> NDArray:
    """Returns an array that joins array b to array a.

    Args:
        a (Tuple[ArrayLike, ArrayLike]): Two arrays to concatenate.

    Returns:
        NDArray: Concatenated array.
    """
    return _numpy.concatenate(a)


def linspace(start: float, stop: float, n: int) -> NDArray:
    """Returns an array of n values evenly spaced over the specified interval.

    Args:
        start (float): Start value.
        stop (float): Stop value.
        n (int): Number of samples.

    Returns:
        NDArray: Evenly spaced values.
    """
    return _numpy.linspace(start, stop, n)


def ones(n: int) -> NDArray:
    """Returns an array of size n filled with ones.

    Args:
        n (int): Size of the array.

    Returns:
        NDArray: Array filled with ones.
    """
    return _numpy.ones(n)


def zeros(n: int) -> NDArray:
    """Returns an array of size n filled with zeros.

    Args:
        n (int): Size of the array.

    Returns:
        NDArray: Array filled with zeros.
    """
    return _numpy.zeros(n)


def flatten(a: ArrayLike) -> NDArray:
    """Returns a copy of an array collapsed into one dimension.

    Args:
        a (ArrayLike): Input array.

    Returns:
        NDArray: Flattened array.
    """
    return _numpy.array(a).flatten()


def reshape(a: ArrayLike, shape: Tuple[int, int]) -> NDArray:
    """Transforms an array to an array of size (n,m).

    Args:
        a (ArrayLike): Input array.
        shape (Tuple[int, int]): Desired shape (rows, columns).

    Returns:
        NDArray: Reshaped array.
    """
    return _numpy.array(a).reshape(shape)


def shape(a: ArrayLike) -> Tuple[int, ...]:
    """Returns the size of the array in the form (n,m).

    Args:
        a (ArrayLike): Input array.

    Returns:
        Tuple[int, ...]: Shape of the array.
    """
    return _numpy.array(a).shape


def tolist(a: ArrayLike) -> List:
    """Converts an array into a list.

    Args:
        a (ArrayLike): Input array.

    Returns:
        List: Converted list.
    """
    return _numpy.array(a).tolist()


def transpose(a: ArrayLike) -> NDArray:
    """Returns a transposed array.

    Args:
        a (ArrayLike): Input array.

    Returns:
        NDArray: Transposed array.
    """
    return _numpy.transpose(a)


def argmax(a: ArrayLike) -> int:
    """Returns the indices of the maximum values.

    Args:
        a (ArrayLike): Input array.

    Returns:
        int: Index of the maximum value.
    """
    return _numpy.argmax(a)


def argmin(a: ArrayLike) -> int:
    """Returns the indices of the minimum values.

    Args:
        a (ArrayLike): Input array.

    Returns:
        int: Index of the minimum value.
    """
    return _numpy.argmin(a)


def dot(a: ArrayLike, b: ArrayLike) -> NDArray:
    """Returns the dot product of two arrays.

    Args:
        a (ArrayLike): First array.
        b (ArrayLike): Second array.

    Returns:
        NDArray: Dot product.
    """
    return _numpy.dot(a, b)


def cross(a: ArrayLike, b: ArrayLike) -> NDArray:
    """Returns the cross product of two arrays.

    Args:
        a (ArrayLike): First array.
        b (ArrayLike): Second array.

    Returns:
        NDArray: Cross product.
    """
    return _numpy.cross(a, b)


def max(a: ArrayLike) -> Union[int, float]:
    """Returns the maximum of the elements.

    Args:
        a (ArrayLike): Input array.

    Returns:
        Union[int, float]: Maximum value.
    """
    return _numpy.max(a)


def min(a: ArrayLike) -> Union[int, float]:
    """Returns the minimum of the elements.

    Args:
        a (ArrayLike): Input array.

    Returns:
        Union[int, float]: Minimum value.
    """
    return _numpy.min(a)


def mean(a: ArrayLike) -> float:
    """Returns the mean of the elements.

    Args:
        a (ArrayLike): Input array.

    Returns:
        float: Mean value.
    """
    return _numpy.mean(a)


def median(a: ArrayLike) -> float:
    """Returns the median of the elements.

    Args:
        a (ArrayLike): Input array.

    Returns:
        float: Median value.
    """
    return _numpy.median(a)


def polyfit(x: ArrayLike, y: ArrayLike, d: int) -> NDArray:
    """Fits a polynomial regression of degree d to the points (x,y).

    Returns a vector of coefficients that minimizes the squared error
    in the order d, d-1, …, 0.

    Args:
        x (ArrayLike): x values.
        y (ArrayLike): y values.
        d (int): Degree of the polynomial.

    Returns:
        NDArray: Polynomial coefficients.
    """
    return _numpy.polyfit(x, y, d)


def polyval(p: ArrayLike, x: ArrayLike) -> NDArray:
    """Evaluates the polynomial p at x.

    Args:
        p (ArrayLike): Polynomial coefficients.
        x (ArrayLike): Values at which to evaluate.

    Returns:
        NDArray: Evaluated values.
    """
    return _numpy.polyval(p, x)


def size(a: ArrayLike) -> int:
    """Returns the number of elements in the array.

    Args:
        a (ArrayLike): Input array.

    Returns:
        int: Number of elements.
    """
    return _numpy.size(a)


def sort(a: ArrayLike) -> NDArray:
    """Sorts the array in ascending order.

    Args:
        a (ArrayLike): Input array.

    Returns:
        NDArray: Sorted array.
    """
    return _numpy.sort(a)


def std(a: ArrayLike) -> float:
    """Returns the standard deviation of the elements.

    Args:
        a (ArrayLike): Input array.

    Returns:
        float: Standard deviation.
    """
    return _numpy.std(a)


def sum(a: ArrayLike) -> Union[int, float]:
    """Returns the sum of the elements.

    Args:
        a (ArrayLike): Input array.

    Returns:
        Union[int, float]: Sum of elements.
    """
    return _numpy.sum(a)



import math as _math
from typing import Tuple

e = _math.e
pi = _math.pi 

def sqrt(x: float) -> float:
    """Square root, type sqrt(x) for √x.

    Args:
        x (float): A non-negative number.

    Returns:
        float: The square root of x.
    """    
    return _math.sqrt(x)

def pow(x: float, y: float) -> float:
    """Power, type pow(x,y) for x ** y.

    Args:
        x (float): Base
        y (float): Exponent

    Returns:
        float:  x ** y.
    """    
    return _math.pow(x, y)

def exp(x: float) -> float:
    """Exponential, type exp(x) for e ** x.

    Args:
        x (float): Exponent

    Returns:
        float: e ** x
    """    
    return _math.exp(x)

def expm1(x: float) -> float:
    """Exponential, type exp(x) for e ** x - 1.

    Args:
        x (float): Exponent

    Returns:
        float: e ** x - 1
    """    
    return _math.exp(x)

def log(x: float) -> float:
    """Natural logarithm: log(x) calculates ln(x).

    Args:
        x (float): Positive number.

    Returns:
        float: ln(x)
    """    
    return _math.log(x)


def log2(x: float) -> float:
    """Base-2 logarithm: log(x) calculates log2(x).

    Args:
        x (float): Positive number.

    Returns:
        float: log2(x).
    """    
    return _math.log2(x)


def log10(x: float) -> float:
    """Base-10 logarithm: log(x) calculates log10(x).

    Args:
        x (float): Positive number.

    Returns:
        float: log10(x).
    """    
    return _math.log10(x)

def cosh(x: float) -> float:
    """Return the hyperbolic cosine of x."""
    return _math.cosh(x)

def sinh(x: float) -> float:
    """Return the hyperbolic sine of x."""
    return _math.sinh(x)

def tanh(x: float) -> float:
    """Return the hyperbolic tangent of x."""
    return _math.tanh(x)

def acosh(x: float) -> float:
    """Return the inverse hyperbolic cosine of x."""
    return _math.acosh(x)

def asinh(x: float) -> float:
    """Return the inverse hyperbolic sine of x."""
    return _math.asinh(x)

def atanh(x: float) -> float:
    """Return the inverse hyperbolic tangent of x."""
    return _math.atanh(x)

def cos(x: float) -> float:
    """Return the cosine of x radians."""
    return _math.cos(x)

def sin(x: float) -> float:
    """Return the sine of x radians."""
    return _math.sin(x)

def tan(x: float) -> float:
    """Return the tangent of x radians."""
    return _math.tan(x)

def acos(x: float) -> float:
    """Return the arc cosine of x in radians."""
    return _math.acos(x)

def asin(x: float) -> float:
    """Return the arc sine of x in radians."""
    return _math.asin(x)

def atan(x: float) -> float:
    """Return the arc tangent of x in radians."""
    return _math.atan(x)

def atan2(y: float, x: float) -> float:
    """Return atan(y / x) in radians, taking into account the quadrant."""
    return _math.atan2(y, x)

def ceil(x: float) -> int:
    """Return the ceiling of x as an int."""
    return _math.ceil(x)

def floor(x: float) -> int:
    """Return the floor of x as an int."""
    return _math.floor(x)

def trunc(x: float) -> int:
    """Return the integer part of x, truncated towards 0."""
    return _math.trunc(x)

def fabs(x: float) -> float:
    """Return the absolute value of x."""
    return _math.fabs(x)

def copysign(x: float, y: float) -> float:
    """Return x with the sign of y."""
    return _math.copysign(x, y)


def fmod(a: float, b: float) -> float:
    """Return the remainder of a / b."""
    return _math.fmod(a, b)

def gcd(a: int, b: int) -> int:
    """Return the greatest common divisor of a and b."""
    return _math.gcd(a, b)

def frexp(x: float) -> Tuple[float,int]:
    """Return the mantissa and exponent of x such that x = mantissa * 2**exponent."""
    return _math.frexp(x)

def ldexp(x: float, i: int) -> float:
    """Return x * (2 ** i)."""
    return _math.ldexp(x, i)

def modf(x: float) -> Tuple[float,float]:
    """Return the fractional and integer parts of x."""
    return _math.modf(x)

def isfinite(x: float) -> bool:
    """Return True if x is finite."""
    return _math.isfinite(x)

def isinf(x: float) -> bool:
    """Return True if x is infinite."""
    return _math.isinf(x)

def isnan(x: float) -> bool:
    """Return True if x is NaN."""
    return _math.isnan(x)

# Conversion
def radians(x: float) -> float:
    """Convert x from degrees to radians."""
    return _math.radians(x)

def degrees(x: float) -> float:
    """Convert x from radians to degrees."""
    return _math.degrees(x)

def erf(x: float) -> float:
    """Return the error function at x."""
    return _math.erf(x)

def erfc(x: float) -> float:
    """Return the complementary error function at x."""
    return _math.erfc(x)

def gamma(x: float) -> float:
    """Return the gamma function at x."""
    return _math.gamma(x)

def lgamma(x: float) -> float:
    """Return the natural logarithm of the absolute value of the gamma function at x."""
    return _math.lgamma(x)

__all__ = [
    "e", "pi", "sqrt", "pow", "exp", "expm1", "log", "log2", "log10",
    "cos", "sin", "tan", "acos", "asin", "atan", "atan2",
    "cosh", "sinh", "tanh", "acosh", "asinh", "atanh",
    "ceil", "floor", "trunc", "fabs", "copysign",
    "fmod", "gcd", "frexp", "ldexp", "modf",
    "isfinite", "isinf", "isnan",
    "radians", "degrees",
    "erf", "erfc", "gamma", "lgamma"
]


import cmath as _cmath
from typing import Tuple

e = _cmath.e
pi = _cmath.pi

def phase(z: complex) -> float:
    """Phase of z in radians, for instance phase(1j)=1.570796326794897.

    Args:
        z (complex): A complex number.

    Returns:
        float: Phase of `z` in radians, between -pi and pi.
    """
    return _cmath.phase(z)

def polar(z: complex) -> Tuple[float, float]:
    """Representation of z in polar coordinates: polar(1j) returns (1.0, 1.570796326794897).

    Args:
        z (complex): A complex number.

    Returns:
        float:  Tuple[float, float]: (modulus, phase)
    """    
    return _cmath.polar(z)

def rect(r: float, phi: float) -> complex:
    """Representation of a complex number in cartesian coordinates.

    Args:
        r (float): Modulus (distance from origin)
        phi (float): Angle in radians

    Returns:
        complex: Complex number in cartesian form
    """    
    return _cmath.rect(r, phi)

def exp(x: complex) -> complex:
    """Exponential function of x.

    Args:
        x (complex): Exponent

    Returns:
        complex: e ** x
    """    
    return _cmath.exp(x)


def log(x: complex) -> complex:
    """Natural logarithm of x.

    Args:
        x (complex): Complex number

    Returns:
        complex: Natural logarithm of x
    """    
    return _cmath.log(x)

def sqrt(x: complex) -> complex:
    """Square root of x.

    Returns the principal square root of x.
    
    Args:
        x (complex): Complex number

    Returns:
        complex: Square root of x
    """
    return _cmath.sqrt(x)

def cos(x: complex) -> complex:
    """Cosine of x (in radians).

    Args:
        x (complex): Complex number in radians

    Returns:
        complex: Cosine of x
    """
    return _cmath.cos(x)

def sin(x: complex) -> complex:
    """Sine of x (in radians).

    Args:
        x (complex): Complex number in radians

    Returns:
        complex: Sine of x
    """
    return _cmath.sin(x)

__all__ = [
    "e", "pi", "phase", "polar", "rect", "exp", "log", "sqrt", "cos", "sin"
]


from typing import Tuple, Union
import turtle as _turtle
import sys

_t: _turtle.Turtle = None
_screen: _turtle.Screen = None

def init(width: int = 320, height: int = 240, title: str = "NumWorks Turtle") -> None:
    global _t, _screen
    _screen = _turtle.Screen()
    _screen.setup(width=width, height=height)
    _screen.getcanvas().winfo_toplevel().protocol("WM_DELETE_WINDOW", sys.exit)
    _screen.title(title)
    _t = _turtle.Turtle()
    
def forward(x: float) -> None:
    """Move forward by x pixels

    Args:
        x (float): Distance to move forward.
    """    
    _t.forward(x)
    
def backward(x: float) -> None:
    """Move backward by x pixels

    Args:
        x (float): Distance to move backward.
    """    
    _t.backward(x)
    
def right(a: float) -> None:
    """Turn right by a degrees.

    Args:
        a (float): Angle in degrees to turn clockwise.
    """    
    _t.right(a)

def left(a: float) -> None:
    """Turn left by a degrees.

    Args:
        a (float): Angle in degrees to turn clockwise.
    """    
    _t.left(a)
    
def goto(x: float, y: float) -> None:
    """Move to (x, y) coordinates.

    Args:
        x (float): X-coordinate.
        y (float): Y-coordinate.
    """    
    _t.goto(x, y)

def setheading(a: float) -> None:
    """Set the orientation by a degrees.

    Args:
        a (float): Angle in degrees
    """    
    _t.setheading(a)

def circle(r: float) -> None:
    """Circle of radius r pixels.

    Args:
        r (float): Radius of the circle in pixels.
    """    
    _t.circle(r)

def speed(x: int) -> None:
    """Drawing speed (x between 0 and 10).

    Args:
        x (int): Speed from fastest to slowest
    """    
    _t.speed(x)
    
def position() -> Tuple[float, float]:
    """Return the current (x,y) location.

    Returns:
        Tuple[float, float]: Current (x, y) coordinates.
    """    
    return _t.position()

def heading() -> float:
    """Return the current heading.

    Returns:
        float: Current heading in degrees.
    """    
    return _t.heading()

def pendown() -> None:
    """Pull the pen down."""    
    _t.pendown()

def penup() -> None:
    """Pull the pen up."""    
    _t.penup()
    
def pensize(x: int) -> None:
    """Set the line thickness to x pixels.

    Args:
        x (int): Pen thickness in pixels.
    """    
    _t.pensize(x)

def write(text: str) -> None:
    """Writes the text placed as an argument at the position of the turtle.

    Args:
        text (str): Text string to display.
    """    
    _t.write(text)

def isdown() -> bool:
    """Return True if the pen is down.

    Returns:
        bool: True if the pen is down, False otherwise.
    """    
    return _t.isdown()

def reset() -> None:
    """Reset the drawing."""
    _t.reset()

def showturtle() -> None:
    """Show the turtle"""
    _t.showturtle()

def hideturtle() -> None:
    """Hide the turtle."""
    _t.hideturtle()
    
def color(c: Union[str, Tuple[int, int, int]]) -> None:
    """Set the pen color.

    Args:
        c (Union[str, Tuple[int, int, int]]): Color name or RGB tuple (0-255).
    """    
    _t.color(c)

def colormode(x: Union[int, float]) -> None:
    """Set the color mode for RGB tuples used in `color()`.

    Args:
        x (int | float): 
            - 1.0 → colors must be specified as floats in the range [0.0, 1.0], e.g., (0.5, 1.0, 0.5)
            - 255 → colors must be specified as integers in the range [0, 255], e.g., (128, 255, 128)
            By default, the color mode is 255.
    """
    _turtle.colormode(x)

def done() -> None:
    """Keep the Turtle window open until manually closed."""
    if _screen:
        _turtle.done()

blue = "blue"
red = "red"
green = "green"
yellow = "yellow"
brown = "brown"
black = "black"
white = "white"
pink = "pink"
orange = "orange"
purple = "purple"
grey = "grey"

__all__ = [
    "forward", "backward", "right", "left", "goto", "setheading",
    "circle", "speed", "position", "heading", "pendown", "penup",
    "pensize", "write", "isdown", "reset", "showturtle", "hideturtle",
    "color", "colormode", "blue", "red", "green", "yellow", "brown", "black", "white", "pink", "orange", "purple", "grey"
]


import time as _time;

def monotonic() -> float:
    """Returns the value of the clock at the time the function is called.
    """    
    return _time.monotonic()

def sleep(t: float) -> None:
    """Pauses execution for t seconds.

    Args:
        t (float): The number of seconds to pause execution
    """    
    _time.sleep(t)
    
__all__ = [
    "sleep",
    "monotonic"
]


import matplotlib.pyplot as _pyplot
from typing import List, Union, Tuple

_fig, _ax = _pyplot.subplots()

def arrow(x: float, y: float, dx: float, dy: float, head_width: float = 0.05, color: str = "black") -> None:
    """
    Draws an arrow from (x, y) to (x+dx, y+dy).

    Args:
        x (float): Starting x-coordinate.
        y (float): Starting y-coordinate.
        dx (float): Change in x (arrow length along x).
        dy (float): Change in y (arrow length along y).
        head_width (float, optional): Width of the arrow head. Defaults to 0.05.
        color (str, optional): Color of the arrow. Defaults to "black".
    """
    _ax.arrow(x, y, dx, dy, head_width=head_width, color=color)

def axis(bounds: Union[Tuple[float, float, float, float], str, bool]) -> None:
    """
    Sets the display window or controls the axes visibility.

    - `axis((xmin, xmax, ymin, ymax))` sets the axes limits.
    - `axis("on")` or `axis(True)` shows the axes.
    - `axis("off")` or `axis(False)` hides the axes.
    - `axis("auto")` resets axes to automatic limits.

    Args:
        bounds (tuple[str, bool]): A 4-tuple (xmin, xmax, ymin, ymax), a string "on"/"off"/"auto",
                                   or a boolean True/False to show/hide axes.
    """
    if bounds in ["on", True]:
        _ax.axis("on")
    elif bounds in ["off", False]:
        _ax.axis("off")
    elif bounds == "auto":
        _ax.axis("auto")
    else:
        _ax.axis(bounds)

def bar(x: List[float], height: List[float], bin_width: float = 0.8, bottom: float = 0, color: str = "black") -> None:
    """
    Draws a bar plot.

    Args:
        x (List[float]): x-coordinates of the bars.
        height (List[float]): Heights of the bars.
        bin_width (float, optional): Width of each bar. Defaults to 0.8.
        bottom (float, optional): Bottom y-value. Defaults to 0.
        color (str, optional): Bar color. Defaults to "black".
    """
    _ax.bar(x, height, width=bin_width, bottom=bottom, color=color)

def grid(on: bool = True) -> None:
    """
    Displays the grid if it is hidden or hides the grid if it is displayed.

    Args:
        on (bool, optional): True to show grid, False to hide. Defaults to True.
    """
    _ax.grid(on)
    
def hist(x: List[float], bins: Union[int, List[float]] = 10, color: str = "black") -> None:
    """
    Plots a histogram using the values in the x list.

    Args:
        x (List[float]): Data to bin.
        bins (int or List[float], optional): Number of bins or bin edges. Defaults to 10.
        color (str, optional): Color of the bars. Defaults to "black".
    """
    _ax.hist(x, bins=bins, color=color)

def plot(x: List[float], y: List[float] = None, color: str = "black") -> None:
    """
    Plots the y list versus the x list

    Args:
        x (List[float]): x-coordinates or y-values if y is None.
        y (List[float], optional): y-coordinates. Defaults to None.
        color (str, optional): Line color. Defaults to "black".
    """
    if y is None:
        y = x
        x = list(range(len(y)))
    _ax.plot(x, y, color=color)
    
def scatter(x: List[float], y: List[float], color: str = "black") -> None:
    """
    Plots the y list versus the x list

    Args:
        x (List[float]): x-coordinates.
        y (List[float]): y-coordinates.
        color (str, optional): Point color. Defaults to "black".
    """
    _ax.scatter(x, y, color=color)
    
def show() -> None:
    """
    Display the figure.
    """
    _pyplot.show()

def text(x: float, y: float, string: str, color: str = "black") -> None:
    """
    Displays the text set as an argument at the (x,y) coordinates.

    Args:
        x (float): x-coordinate.
        y (float): y-coordinate.
        string (str): Text to display.
        color (str, optional): Text color. Defaults to "black".
    """
    _ax.text(x, y, string, color=color)

